{-# OPTIONS -cpp #-}

module Interface.Interactive where

import Control.Monad.Error
import Control.Monad.Reader
import Data.Set
import System.Console.Readline
import Text.PrettyPrint

import Syntax.Parser
import Syntax.ParseMonad    ( ParseResult(..), showParseFailure )
import Syntax.Pretty
import Syntax.ConcreteToInternal

import TypeChecker
import TypeChecker.Monad
import TypeChecker.Monad.Definitions
import TypeChecker.Monad.Options
import TypeChecker.Computation
import TypeChecker.UnfoldConstants
import TypeChecker.Terms

import Utils.Map
import Utils.Monads

-- | The interaction loop.
interactionLoop :: TC () -> TC ()
interactionLoop typeCheck = load >> loop
    where
	load =
	    typeCheck `catchError`
		\e -> liftIO $ do   print e
				    putStrLn "Failed."

	loop =
	    do	ms <- liftIO $ readline "Main> "
		case ms of
		    Nothing	-> fail "Unexpected readline failure"
		    Just ":q"	-> return ()
                    Just ":?"   -> 
                        do  liftIO $ addHistory ":r"
			    liftIO $ putStr help
                            loop
		    Just ":r"	->
			do  liftIO $ addHistory ":r"
			    defs <- load
			    loop
		    Just s	-> do	liftIO $ addHistory s
					evalTerm s
					loop
	    `catchError` \e ->
		do  liftIO $ print e
		    loop

evalTerm s =
    do	case parse termParser s of
	    ParseFailed p why what  -> fail $ showParseFailure p why what
	    ParseOk _ ct		->
		do  defs	<- getDefinitions
		    let t = toInternal defs ct
		    t' <- inferType t
		    w  <- unfoldAllConstants <$> evaluateTypedTerm t'
		    v  <- termType t'
		    liftIO $ print $
			    sep [ pretty w
				, nest 2 (text "::" <+> pretty v)
				]
		where
		    toInternal defs ct =
			runReader
			    (termToInternalM ct)
			    (Syntax.ConcreteToInternal.emptyEnv
				{ eDefs	    = mkSet $ definedNames defs
				, eFixities = emptyMap    -- TODO
				}
			    )

definedNames defs   = keysMap defs

-- | The logo that prints when agdaLight is started in interactive mode.
splashScreen :: String
splashScreen = unlines
    [ "               _       _             _"
    , "              | |     | |           | |    _"
    , "              | |     | |    _      | |   | |_ _"
    , "  ___  ___  __| | ___ | |   (_) ___ | |___|  _(_)"
    , " / _ \\/ _ \\/ _  |/ __\\| |   | |/ _ \\|  _  \\ | | | agdaLight Interactive"
    , "/ /_|/ /_\\ \\/_| / /_| | |___| / /_\\ \\ | | | |_| |   version " ++ _VERSION_
    , "\\____\\___  /____\\_____|_____|_\\___  /_| |_|__/|_| Type :? for help."
    , "      __/ /                    __/ /"
    , "      \\__/                     \\__/"
    ]

-- | The help message
help :: String
help = unlines
    [ "Command overview"
    , ":q    Quit."
    , ":?    Help (this message)."
    , ":r    Reload input files."
    , "<exp> Infer type of agdaLight expression <exp> and evaluate it."
    ]

