{-# OPTIONS -cpp -fglasgow-exts #-}

module Interface.Interactive where

import Prelude hiding (print, putStr, putStrLn)
import IO.Output

import Control.Monad.Error
import Control.Monad.Reader
import Data.Char
import Data.Set as Set
import Data.Map as Map
import Data.List as List
import Data.Maybe
import Text.PrettyPrint

import Syntax.Position
import Syntax.Name
import qualified Syntax.Internal as I
import qualified Syntax.Concrete as C
import Syntax.Parser
import Syntax.ParseMonad
import Syntax.Pretty
import Syntax.ConcreteToInternal
import Syntax.Values

import TypeChecker
import TypeChecker.Monad
import TypeChecker.Monad.Errors
import TypeChecker.Monad.Definitions
import TypeChecker.Monad.Datatypes
import TypeChecker.Monad.Options
import TypeChecker.Monad.MetaVars
import TypeChecker.Values
import TypeChecker.Values.Refresh
import TypeChecker.UnfoldConstants
import TypeChecker.MetaVars
import TypeChecker.Values.DeBruijn
import TypeChecker.Errors

import Compilers.Run

import Interface.ChaseImports
import Interface.Options

import IO.Readline

import Utils.Map as Map
import Utils.Monads
import Utils.Names

data ExitCode a = Continue | ContinueIn Environment | Return a

type Command a = (String, [String] -> TC (ExitCode a))

matchCommand :: String -> [Command a] -> Either [String] ([String] -> TC (ExitCode a))
matchCommand x cmds =
    case List.filter (isPrefixOf x . fst) cmds of
	[(_,m)]	-> Right m
	xs	-> Left $ List.map fst xs

interaction :: String -> [Command a] -> (String -> TC (ExitCode a)) -> TC a
interaction prompt cmds eval = loop
    where
	go (Return x)	    = return x
	go Continue	    = loop
	go (ContinueIn env) = local (const env) loop

	loop =
	    do	ms <- liftIO $ readline prompt
		case fmap words ms of
		    Nothing		  -> return $ error "** EOF **"
		    Just []		  -> loop
		    Just ((':':cmd):args) ->
			do  liftIO $ addHistory (fromJust ms)
			    case matchCommand cmd cmds of
				Right c	-> go =<< c args
				Left []	->
				    do	liftIO $ putStrLn $ "Unknown command '" ++ cmd ++ "'"
					loop
				Left xs	->
				    do	liftIO $ putStrLn $ "More than one command match: " ++ concat (intersperse ", " xs)
					loop
		    Just _ ->
			do  liftIO $ addHistory (fromJust ms)
			    go =<< eval (fromJust ms)
	    `catchError` \e ->
		do  liftIO $ print e
		    loop

-- | The interaction loop.
interactionLoop :: TC ScopeEnv -> TC ()
interactionLoop typeCheck =
    do	env  <- reload
	it   <- fresh "it"
	defs <- compileDefs Map.empty
	withScopeEnv env $ interaction "Main> " (commands it defs) (evalTerm it)
    where
	reload = typeCheck `catchError`
		    \e -> liftIO $ do print e
				      putStrLn "Failed."
				      return emptyScopeEnv

	commands it defs =
	    [ "quit"	|>  \_ -> return $ Return ()
	    , "help"	|>  \_ -> do liftIO $ putStr help
				     return Continue
	    , "?"	|>  \_ -> do liftIO $ putStr help
				     return Continue
	    , "type"	|>  \args ->
				do  typeTerm it (unwords args)
				    return Continue
	    , "exec"	|>  \args ->
				do  (v,_) <- inferTerm' it $ unwords args
				    w	  <- compileValue Map.empty v
				    liftIO $ print $ evaluate defs w
				    return Continue
	    , "reload"	|>  \_ -> do scope <- reload
				     env <- withScopeEnv scope ask
				     return $ ContinueIn env
	    , "load"	|>
		\args -> case args of
		    [file]  -> do opts <- commandLineOptions
				  files <- liftIO $ chaseImports (optIncludeDirs opts) [file]
				  liftIO $
				    do	putStrLn "After import chasing:"
					mapM_ (\file -> putStrLn $ "  " ++ file) files
				  env <- ask
				  let env'  = env { envOptions = opts { optInputFiles = files } }
				  senv <- local (const env') reload
				  return $ ContinueIn $ env' { envScopeEnv = senv }
		    _	    -> do liftIO $ putStrLn "Usage :load file_name"
				  return Continue
	    ]
	    where
		(|>) = (,)

parseTerm :: ScopeEnv -> String -> TC (I.Term, Maybe I.Term)
parseTerm = parseTermAt (startPos "")

parseTermAt :: Posn -> ScopeEnv -> String -> TC (I.Term, Maybe I.Term)
parseTermAt pos scope s =
    case parse' defaultParseFlags
		pos 0 interactParser s of
	ParseFailed p why what  -> fail $ showParseFailure p why what
	ParseOk _ ct		->
	    do  used <- getUsedNames
		case toInternal used scope ct of
		    Left err -> sayWhen whenCheckingScope $ fail err
		    Right t  -> return t
    where
	toInternal :: UsedNames -> ScopeEnv -> Either C.Term C.Typing ->
					       Either String (I.Term, Maybe I.Term)
	toInternal used scope (Left ct) =
	    runConvM' used scope $
		do  t <- termToInternalM ct
		    return (t, Nothing)
	toInternal used scope (Right (C.Typing ctm ctp)) =
	    runConvM' used scope $
		do  tm <- termToInternalM ctm
		    tp <- termToInternalM ctp
		    return (tm, Just tp)

inferTerm it s = allowUnsolvedMetas $ inferTerm' it s

inferTerm' it s =
    do	env <- addDefinedName it <$> getScopeEnv
	(tm,tp) <- parseTerm env s
	case tp of
	    Nothing -> refresh =<< inferType tm
	    Just tp -> do t <- isType tp
			  v <- checkType tm t
			  refresh (v,t)

typeTerm it s =
    do	(_,t) <- inferTerm it s
	liftIO $ print $ sep [ text s <+> text ":"
			     , nest 2 $ pretty t
			     ]

evalTerm it s =
    do	(w,wt)  <- inferTerm it s
	w'	<- unfoldAllConstants w
	wt'	<- typeMapM unfoldAllConstants wt
	addDefinition it $ DefVal 0 0 0 (ConcrDef [ClauseVal [] w']) wt'
	(w'',wt'')  <- unBruijn (w',wt')
	liftIO $ print $
		sep [ pretty w''
		    , nest 2 (text ":" <+> pretty wt'')
		    ]
	return Continue

definedNames defs   = Map.keys defs
constructors dats   = Prelude.map fst $ concat $ Map.elems dats
datatypes dats = Map.keys dats

-- | The logo that prints when agdaLight is started in interactive mode.
splashScreen :: String
splashScreen = unlines
    [ "               _       _             _"
    , "              | |     | |           | |    _"
    , "              | |     | |    _      | |   | |_ _"
    , "  ___  ___  __| | ___ | |   (_) ___ | |___|  _(_)"
    , " / _ \\/ _ \\/ _  |/ __\\| |   | |/ _ \\|  _  \\ | | | agdaLight Interactive"
    , "/ /_|/ /_\\ \\/_| / /_| | |___| / /_\\ \\ | | | |_| |   version " ++ _VERSION_
    , "\\____\\___  /____\\_____|_____|_\\___  /_| |_|__/|_| Type :? for help."
    , "      __/ /                    __/ /"
    , "      \\__/                     \\__/"
    ]

-- | The help message
help :: String
help = unlines
    [ "Command overview"
    , ":quit         Quit."
    , ":help or :?   Help (this message)."
    , ":reload       Reload input files."
    , ":load file    Load file."
    , ":type e       Infer the type of e"
    , "<exp> Infer type of agdaLight expression <exp> and evaluate it."
    ]

