{-# OPTIONS -fglasgow-exts #-}

module TypeChecker.UnfoldConstants where

import Prelude hiding	    ( print, putStr, putStrLn )
import IO.Output

import Control.Monad.Trans
import Control.Monad
import Data.Monoid

import Syntax.Internal
import Syntax.Values

import TypeChecker.Context
import TypeChecker.Values
import TypeChecker.Values.Refresh
import TypeChecker.Abstractions
import TypeChecker.Patterns.Match
import TypeChecker.Monad
import TypeChecker.Monad.Definitions
import TypeChecker.Monad.Rewriting
import TypeChecker.Monad.Context

import Utils.Monads
import Utils.Pairs
import Utils.Names

import Syntax.Pretty

-- Unfolding constants ----------------------------------------------------

data UnfoldStrategy = TopLevel | Recursive

-- | Unfolds top level constants. Calls 'genUnfoldConstApp' with 
--   a stop condition that yells as soon as it sees something
--   that isn't a constant.
unfoldConstants :: Value -> TC Value
unfoldConstants v =
    genUnfoldConstants TopLevel (\_ _ _ -> False) (const Nothing) v

-- | Unfold all constants everywhere.
unfoldAllConstants :: Value -> TC Value
unfoldAllConstants v =
    genUnfoldConstants Recursive (\_ _ _ -> False) (const Nothing) v

-- | Top level unfolding of everything that's not one of the
--   listed constants.
unfoldSomeConstants :: [Name] -> Value -> TC Value
unfoldSomeConstants nope =
    genUnfoldConstants TopLevel (\x _ _ -> x `elem` nope) (const Nothing)

-- | Recursive unfolding of everything that's not one of the
--   listed constants.
unfoldAllButSomeConstants :: [Name] -> Value -> TC Value
unfoldAllButSomeConstants nope =
    genUnfoldConstants Recursive (\x _ _ -> x `elem` nope) (const Nothing)

-- | The most general constant unfolding function.
--   Takes a flag indicating if we should only unfold top-level
--   and a don't unfold me condition.
genUnfoldConstants ::
    UnfoldStrategy		->  -- TopLevel or Recursive
    (Name -> DefValue -> [AppOrProj] -> Bool)
				->  -- don't unfold me
    (Value -> Maybe Value)	->  -- custom unfold rules
    Value -> TC Value
genUnfoldConstants strategy nounfold special v
    | Just v' <- special v  = recCall v'
    | otherwise		    =
	--liftIO (print $ text "unfold" <+> pretty v) >>
    do  v' <- refreshTop v
	case vAppProjView v' of
	    VAPVDef x h vps ->
		do  d@(DefVal _ ar' fv def _) <- lookupDef x
		    let cs = getClausesTC def
			ar = ar' + fv
			n  = length (takeWhile isApp vps)
			res | nounfold x d vps  = vAppOrProjs (vDef x h)
						    <$> mapM recCallOrId vps
			    | n < ar		= vAppOrProjs (vDef x h)
						    <$> mapM recCallOrId vps
			    | otherwise		=
				do  let (vs0,vs1,vps0) = split vps
				    (vs0',m) <- match cs vs0
				    case m of
					Nothing ->
					    do	vps' <- mapM recCallOrId
							$ map IsApp (vs0' ++ vs1) ++ vps0
						return $ vAppOrProjs (vDef x h) vps'
					Just v' ->
					    recCall_ $ vAppOrProjs
						       (vApps v' vs1) vps0
				where
				    split vps = (vs0,vs1,vps0)
					where
					    (vs,vps0) = map unApp -*- id
							$ splitAt n vps
					    (vs0,vs1) = splitAt ar vs
		    r	<- res
		    mr	<- applyRewriting r
		    case mr of
			Left v	-> return v
			Right w -> recCall_ w
		where
		    isApp (IsApp _) = True
		    isApp _	    = False
		    unApp (IsApp v) = v
		    unApp _	    = error "impossible: TypeChecker.UnfoldConstants.unApp"

	    VAPVId x vs		-> vAppOrProjs (vId x) <$> mapM recCallOrId vs
	    VAPVBId x vs	->
		do  mr <- applyRewriting v'
		    case mr of
			Left v	->
			    case vAppProjView v of
				VAPVBId x vs -> vAppOrProjs (vBId x) <$> mapM recCallOrId vs
				_	     -> error "impossible: TypeChecker.UnfoldConstants.unfold (VAPVBId _ _)"
			Right r	-> recCall_ r
	    VAPVCon x h ar vs	-> vAppOrProjs (vCon x h ar) <$> mapM recCallOrId vs
	    VAPVLam cl		-> vLam <$> unfoldAbs cl
	    VAPVPi v cl		-> vPi <$> recCallT v
				       <*> unfoldAbsT v cl
	    VAPVSigma w cl	-> vSigma <$> recCallT w
					  <*> unfoldAbsT w cl
	    VAPVPair v w	-> vPair <$> recCall v
					 <*> recCall w
	    VAPVStruct xs	-> vStruct <$> mapM (\ (x,(h,v)) -> (,) x . (,) h <$> recCall v) xs
	    VAPVMeta _ _	-> return v'
	    VAPVTop		-> return v'
	    VAPVSort _		-> return v'
	    
    where
	recCallOrId (IsApp v)	    = IsApp <$> recCall v
	recCallOrId vp@(IsProj _)   = return vp

	unfoldAbs f =
	    absMapM (\v ->
			do  x <- fresh $ absName f
			    extendContext x (Abstract $ error "oops")
				$ recCall v
		    ) f
	unfoldAbsT t f =
	    absMapM (\v ->
			do  x <- fresh $ absName f
			    extendContext x (Abstract t)
				$ recCallT v
		    ) f

	recCall = case strategy of
		    TopLevel	-> return
		    Recursive	-> recCall_

	recCall_ = genUnfoldConstants strategy nounfold special

	recCallT t = typeMapM recCall t

	vAppOrProjs e vs    = foldl vAppOrProj e vs

	match [] vs = return (vs, Nothing)
	match (ClauseVal ps w : cs) vs =
	    do	-- ? Are there any pitfalls here ?
		(vs',ms) <- unzip <$> zipWithM patternMatch ps vs
		case mconcat ms of
		    MisMatch	-> match cs vs'
		    DontKnow	-> return (vs', Nothing)
		    Match ws	-> return (vs', Just $ vApps w ws)

